Kulitta Program
Donya Quick
Last modified: 18-June-2014

This module provides examples of Kulitta's output using an
interactive interface. The program can be compiled by:

ghc -O2 Kulitta.lhs

Running "Kulitta" will start the program in GUI/MUI mode. 
The program will also respond to different arguments:

Kulitta help		Explains how to use the program.
Kulitta about		More information about the program.
Kulitta basic		Interactive, command-line version. 

To run Kulitta in GHCI, you can load this file and run the
"main" function directly, starting the progam in GUI mode.
However, you will not be able to use the argument-based 
methods of interaction through GHCI. If you would like to
use enter the parameters using a text-based series of 
prompts in GHCI, change the contents of "settings.txt" to  
be "basic" instead of "mui" (which is the default).

For more examples of how to use Kulitta's modules, see those 
in Examples.lhs.

> {-# LANGUAGE Arrows #-}

> module Main where
> import Euterpea hiding (C, D, P, Phrase)
> import JazzFG
> import ClassicalFG
> import GrammarMonad 
> import GrammarRules
> import System.Random
> import PostProc
> import Learning
> import PCFGtoPTGG
> import Rohrmeier hiding (S)
> import ChordSpaces hiding (i)
> import Data.List
> import RuleUtils
> import System.Environment
> import System.IO
> import System.Directory
> import BachStat
> import Euterpea.Experimental
> --import Euterpea.IO.MUI.UISF
> import Play2
> import PlayK
> import Search
> import Constraints
> import FRP.UISF.UISF

> channelOffset = 3 -- this is useful for some synthesizers

Data type definitions to allow the user to specify Kulitta's behavior.

> data Style = Chorale | -- in style of JS Bach
>              JazzChorale | -- grammar, jazz chords, chorale foreground
>              WeirdChorale | -- grammar through OPT-space
>              JazzChords | -- grammar with jazz foreground
>              BossaNova -- grammar with bossa foreground
>    deriving (Eq, Show, Ord, Read)
> data Form = Phrase | AABA
>    deriving (Eq, Show, Ord, Read)
> data GramType = HandBuilt | Learned
>    deriving (Eq, Show, Ord, Read)
> data Info = Info {style :: Style, form :: Form, gram :: GramType, mode :: Mode, lets :: Bool, randKey :: Bool}

=======================

MAIN PROGRAM DEFINITION

Main program description. There are two modes: interacive and automated with arugments.

> main = do
>     hSetBuffering stdout NoBuffering -- required to make things print in the right order
>     putStrLn "\n\n===== Kulitta =====\n"
>     args <- getArgs
>     if length args == 0 then putStrLn "\nHello! The graphical interface is now active. \n" >> 
>                              runDefault else 
>         if args !! 0 == "basic" then runBasicVersion else
>         if args !! 0 == "help" then printDirections else
>         if args !! 0 == "about" then printAbout else
>         if length args < 7 then putStrLn "\nSorry, I couldn't understand those arguments.\n" >>
>                                 printDirections
>         else processArgs args

> runDefault = do
>    files <- getDirectoryContents ""
>    if not (elem "settings.txt" files) then putStrLn "File settings.txt not found." >> runMuiVersion else do
>         startMode <- readFile "settings.txt"
>         if take 5 startMode == "basic" then runBasicVersion else runMuiVersion

> runMuiVersion = mui >> putStrLn "\nGraphical interface closed. Goodbye!\n\n" 
> runBasicVersion = interactive >> putStrLn "\nGoodbye!\n\n"

=======================

MUI DEFINITION

> styles = [Chorale, JazzChorale, WeirdChorale, JazzChords, BossaNova]
> forms = [Phrase, AABA]
> grams = [HandBuilt, Learned]
> modes = [Major, Minor]

> mui = runMUI defaultMUIParams{uiSize=(500,550), uiTitle="Kulitta - Graphical Interface"} $ proc _ -> do
>   label "Please specifty composition parameters for Kulitta." -< ()
>   label "The command prompt will show the generation/playback status." -< ()
>   label "" -< ()
>   (style', mo) <- inputPaneA -< ()
>   (form',gram',mode', instr', let', key') <- inputPaneB -< ()
>   seed' <- leftRight $ label "Random seed: " >>> textboxE "0" -< Nothing
>   outFile <- leftRight $ label "Output File: " >>> textboxE "test.mid" -< Nothing
>   label "" -< ()
>   let s = styles !! style'
>       f = forms !! form'
>       g = grams !! gram'
>       m = modes !! mode'
>       iMIDI = instr'==0
>       iLet = let'==0
>       iKey = key'==0
>       iVal = Info s f g m iLet iKey
>       seed = let x = reads seed' :: [(Int, String)] 
>              in  if null x then 0 else fst $ head x 
>   (g,p) <- buttons -< ()
>   let g' = fmap (const (iVal, seed, outFile, iMIDI)) g
>       p' = fmap (const (outFile, mo)) p
>   basicIOWidget genWrap -< g'
>   basicIOWidget playWrap -< p'
>   returnA -< ()

> genWrap (i, seed, outFile, inst) = automated i seed outFile inst
> playWrap (fname, devID) = do
>    putStrLn "\nPlaying...(please wait)\n" 
>    playX fname devID channelOffset -- previously used play'
>    putStrLn "\nDone!\n\n"

> buttons = leftRight $ proc _ -> do
>   genButton <- edge <<< button "Generate MIDI File" -< ()
>   playButton <- edge <<< button "Play MIDI File" -< ()
>   returnA -< (genButton, playButton)

> inputPaneA = leftRight $ proc _ -> do
>   style <- topDown $ title "Style" $ radio (map show styles) 0 -< ()
>   mo <- selectOutput -< ()
>   returnA -< (style, mo)

> inputPaneB = leftRight $ proc _ -> do
>   (form, gram, lets) <- inputPane2 -< ()
>   (mode, instr, key) <- inputPane3 -< ()
>   returnA -< (form, gram, mode, instr, lets, key) where
>     inputPane2 = topDown $ proc _ -> do
>       form <- topDown $ title "Form" $ radio (map show forms) 0 -< ()
>       gram <- topDown $ title "Harmony Model" $ radio (map show grams) 0 -< ()
>       lets <- topDown $ title "Use Lets" $ radio ["Yes", "No"] 1 -< ()
>       returnA -< (form, gram, lets)
>     inputPane3 = topDown $ proc _ -> do
>       mode <- topDown $ title "Mode" $ radio ["Major", "Minor"] 0 -< ()
>       instr <- topDown $ title "Assign MIDI instruments?" $ radio ["Yes", "No"] 0 -< ()
>       key <- topDown $ title "Random key?" $ radio ["Yes", "No"] 0 -< ()
>       returnA -< (mode, instr, key)

> basicIOWidget :: (a -> IO ()) -> UISF (SEvent a) ()
> basicIOWidget = (>>> arr (const ())) . uisfSinkE



=======================

CONSOLE PROGRAM DEFINITION

> printAbout = do
>     putStrLn "Created by Donya Quick at Yale University (donya.quick@yale.edu)" 
>     putStrLn "For more information, go to http://www.donyaquick.com and click on "
>     putStrLn "Current Research. Relevant publications can be found on the "
>     putStrLn "Yale Haskell Group's website, http://haskell.cs.yale.edu."

> printDirections = do
>     putStrLn "To call Kulitta with a graphical interface, just run 'Kulitta' (no arguments)."
>     putStrLn "To use Kulitta from the command prompt, run 'Kulitta basic' and follow the prompts.\n"
>     putStrLn "To provide arguments, use 'Kulitta s f m g i b x' where"
>     putStrLn "  s = Chorale | JazzChorale | WeirdChorale | JazzChords | BossaNova"
>     putStrLn "  f = Phrase | AABA"
>     putStrLn "  m = Major | Minor"
>     putStrLn "  g = HandBuilt | Learned"
>     putStrLn "  i = an integer, like 392"
>     putStrLn "  b (assign MIDI instruments) = Yes | No"
>     putStrLn "  x = a file path, like 'foo.mid'\n"
>     putStrLn "Run 'Kulitta about' for more information on the program."
>     putStrLn "You can use Ctrl+C to exit at any time while Kulitta is running.\n\n"

> processArgs strs = do
>     let style = read (strs !! 0)
>         form = read (strs !! 1)
>         mode = head $ parseMode (strs !! 2)
>         gram = read (strs !! 3)
>         seed = read (strs !! 4)
>         inst = strs !! 5 == "Yes" || strs !! 5 == "yes"
>         outFile = strs !! 6
>     automated (Info style form gram mode False True) seed outFile inst

> interactive = do 
>     style <- getStyle
>     form <- getForm
>     mode <- getMode
>     gram <- getGram
>     seed <- getSeed
>     inst <- getInstr
>     --useLets <- getLets (gram == HandBuilt)
>     --randomKey <- getKey
>     outFile <- getFilePath
>     automated (Info style form gram mode False True) seed outFile inst

> automated (Info style form gram mode lets key) seed outFile inst = do
>     putStrLn ("\nI will now write a "++ show mode ++ " "++ show style ++
>               " in "++show form++" form with random seed "++ show seed ++ 
>               " and a "++show gram++" grammar, and I will "++
>               " and write it to the file '"++ outFile ++"'.\n")
>     putStrLn "Please be patient - some styles can take a while to write!\n"
>     putStrLn "Working...\n"
>     m <- makePiece (mkStdGen seed) (Info style form gram mode lets key) inst
>     writeMidi outFile m
>     putStrLn ("Done! Please check "++outFile++" to hear what I wrote.\n")

> getStyle = do
>   putStrLn ("I can write the following styles:\n Chorale \t JazzChorale "++
>              "\t WeirdChorale \n JazzChords \t BossaNova ")
>   putStrLn "\nWhat would you like me to write?\n"
>   putStr "Style: "
>   styleStr <- getLine
>   let style = reads styleStr :: [(Style, String)]
>   if null style then putStrLn ("\nSorry, I don't understand. "++
>                             "Please type the style exactly. ") >> getStyle 
>       else return (fst $ head $ style)

> getForm = do
>   putStrLn "\nI can write two forms: Phrase or AABA. Which would you like?\n"
>   putStr "Form: "
>   formStr <- getLine
>   let form = reads formStr :: [(Form, String)]
>   if null form then putStr ("\nSorry, I don't understand. "++
>                            "Plase type only 'Phrase' or 'AABA'. ") >> getForm 
>       else return (fst $ head $ form)

> getMode = do
>   putStrLn "\nShould this be a Major or a Minor piece?\n"
>   putStr "Mode: "
>   modeStr <- getLine
>   let mode = parseMode modeStr
>   if null mode then putStr ("\nSorry, I don't understand. "++
>                     "Please type only 'Major' or 'Minor'. ") >> getMode 
>       else return (head mode) 

> parseMode x = let m = take 5 x in
>       if m=="Major" then [Major] else if m=="Minor" then [Minor] else []

> getSeed = do
>   putStr "\nGive me a random number seed to use: "
>   seedStr <- getLine
>   let seed = reads seedStr :: [(Int, String)]
>   if null seed then putStr ("\nSorry, I don't understand. "++
>                            "Please enter an integer (Int) value. ") >> getSeed
>       else return (fst $ head $ seed)

> getGram = do
>   putStrLn "\nShould I use the HandBuilt or Learned grammar for harmony? "
>   putStr "\nGrammar: "
>   gramStr <- getLine
>   let gram = reads gramStr :: [(GramType, String)]
>   if null gram then putStr ("\nSorry, I don't understand. "++
>                              "Please enter 'HandBuilt' or 'Learned'. ") >> getGram
>       else return (fst $ head $ gram)

> getFilePath = do
>   putStrLn "\nWhere would you like me to write the MIDI file and what should I call it?\n"
>   putStr "Output file path: "
>   filePath <- getLine
>   return $ stripQuotes filePath

> getInstr = do
>   putStrLn "\nShould I assign MIDI instruments?\n"
>   putStr "Yes or No: "
>   instStr <- getLine
>   let inst = if instStr=="Yes" || instStr=="yes" then [True] else
>              if instStr=="No" || instStr=="no" then [False] else []
>   if null inst then putStr ("\nSorry, I don't understand. "++
>                     "Please type only 'Yes' or 'No'. ") >> getInstr 
>       else return (head inst) 


> stripQuotes = filter (not . (`elem` "\"'"))

===============================

ABSTRACT PHRASE GENERATION

This function creates chorale phrases using a hand-built grammar. 

> makeRPhraseH :: StdGen -> Mode -> (Dur,Dur) -> Bool -> Int -> Dur -> Bool -> IO (Constraints, [RChord])
> makeRPhraseH g m (minD,maxD) lets iters len partB= 
>     let ctxt = m -- if m==Major then 0 else 12 
>         tVal = doGen (ruleSet minD lets) iters g ctxt (i len) maxD
>         kVal = findInds [] tVal
>     in  return (kVal, toChords (expand [] tVal) m)

This files store the results of training Kulitta on a Bach chorale corpus.

> minorProbsFile = "minorProbs.txt"
> majorProbsFile = "majorProbs.txt"
> pcfgFile = "pcfg.txt"

These production probabilities can be used to make other phrases. Since we
require that phrases end on one, phrases will be searched until one is found 
that does end on I (ending on I is not guaranteed by the learned grammar).
 
> makeRPhraseB :: StdGen -> Mode ->(Dur,Dur) -> Int -> Dur -> Bool -> IO (Constraints, [RChord])
> makeRPhraseB g m (minD,maxD) iters len partB = do
>     let [g1, g2, g'] = take 3 $ splitN g
>     (startSym, rulesPCFG) <- readPCFG pcfgFile
>     let theFile = if m==Major then majorProbsFile else minorProbsFile
>     probs <- readProbsFinal theFile
>     let avgProbs = map average probs
>         prules = zip avgProbs rulesPCFG
>         rules = toPTGG3 minD prules
>         ctxt = m -- if m==Major then 0 else 12 
>         tPhrase = doGen rules iters g1 ctxt (i len) maxD
>         endsOnI = endingType tPhrase
>         rPhrase = toChords (expand [] $ tPhrase) m
>         rPhrase' = expandTSD2 (tsdSpace' m) (okRTrans m) g2 rPhrase
>     if endsOnI then return ([], rPhrase') else makeRPhraseB g' m (minD,maxD) iters len partB where
>     endingType = (==I) . last . map (\(a,b,c) -> c) . flip toChords Major . expand []

> average xs = sum xs / fromIntegral (length xs)


The doGen function adds a maximum duration constraint to get a more consistent 
distribution of durations. No more than searchLimit generative steps will be tried
to avoid infinite loops caused by generative fixed points.

> searchLimit = 1000

> doGen :: (Eq a) => [Rule a] -> Int -> StdGen -> Mode -> Term a -> Dur -> Term a
> doGen theRules i g ctxt t maxDur = recCheck i (gen'' theRules g ctxt t) where
>     recCheck i ts = let tc = ts !! i in
>         if (goodDurs tc && goodDurs' tc) || i>searchLimit then tc else recCheck (i+1) ts
>     goodDurs :: Predicate (Term a)
>     goodDurs (Let x a e) = goodDurs a && goodDurs e
>     goodDurs (Var x) = True -- can't tell duration here, so assume ok
>     goodDurs (GrammarMonad.S xs) = and $ map goodDurs xs
>     goodDurs (Mod m x) = goodDurs x
>     goodDurs (NT (Chord d c)) = d <= maxDur

> goodDurs' :: Predicate (Term a)
> goodDurs' t = let x = getDurs t in 
>     length (filter (>=hn) x) < length x `div` 2

> getDurs :: Term a -> [Dur]
> getDurs t@(Let x a e) = getDurs (expand [] t)
> getDurs (Var x) = error "getDurs can't handle variables."
> getDurs (GrammarMonad.S xs) = concatMap getDurs xs
> getDurs (Mod m x) = getDurs x
> getDurs (NT (Chord d c)) = [d]

This version of TSD-space is altered from the originally-learned statistics to
improve Kulitta's performance (the analysis had some known problems). 

> tsdSpace' :: Mode -> QSpace CType
> tsdSpace' m = [f I 0 ++ f III 2 ++ f VI 5,
>                f IV 3 ++ f II 1,
>                f V 4 ++ f VII 6] where
>                             --  I   II  III  IV   V   VI  VII
>     rCounts = if m==Major then [34, 13,  4,  11,  25,  7,  3]
>               else             [47,  2,  4,  18,  20,  3,  3]
>     f x i = take (rCounts !! i) $ repeat x


===============================

FOREGROUND GENERATION

The following code configures Kulitta in different ways to generate different 
styles of music according to the user's specifications.

> makePiece g i@(Info s f gr m l k) b = do
>     let [gStruct, gFG] = take 2 $ splitN g
>         genVals = if chorale s then (5, qn, hn, 4) else (5, wn, wn, 8)
>     absStructs <- makeStructure gStruct i genVals
>     theMusic <- makeMusic gFG i absStructs
>     return$ procInstrs b theMusic

> procInstrs :: Bool -> Music a -> Music a
> procInstrs True m = m
> procInstrs False (m1 :=: m2) = procInstrs False m1 :=: procInstrs False m2
> procInstrs False (m1 :+: m2) = procInstrs False m1 :+: procInstrs False m2
> procInstrs False (Modify (Instrument i) m) = procInstrs False m
> procInstrs False m = m

> makeSubStruct gAbs i@(Info s f gr m l k) (iters, minD, maxD, len) partB = do
>    (cons, abs) <- if gr==HandBuilt then makeRPhraseH gAbs m (minD,maxD) l iters len partB else
>                   makeRPhraseB gAbs m (minD,maxD) (iters+1) len partB
>    return (cons, abs)      

> makeStructure gStruct i@(Info s f gr m l k) genVals = do
>     let (g1, g2) = split gStruct
>     structs <- if f==Phrase then sequence [makeSubStruct g1 i genVals False]
>                else sequence [makeSubStruct g1 i genVals False, 
>                               makeSubStruct g2 i genVals True]
>     return structs

> chorale s = elem s [Chorale, JazzChorale, WeirdChorale]

> makeMusic g i@(Info s f gr m l k1) absStructs = do
>     let (k2, gFG) = randomR (0,11::Int) g
>         k = if k1 then k2 else 0
>     let fg = case s of 
>                  Chorale -> addVolume 127 $ buildChorale gFG absStructs (k,m)
>                  JazzChorale -> addVolume 127 $ buildJChorale gFG absStructs (k,m)
>                  WeirdChorale -> addVolume 127 $ buildWChorale gFG absStructs (k,m) 
>                  JazzChords -> addVolume 127 $ buildJazzChords gFG absStructs (k,m)
>                  BossaNova -> buildBossaNova gFG absStructs (k,m)
>     putStrLn ("Key of piece: "++ showKey k m ++"\n")
>     return fg where
>     showKey k m = (["C","C-sharp","D","E-flat","E","F","F-sharp","G","A-flat","A","B-flat","B"] !! k) ++ " " ++ show m

A chorale is pretty straightforward, using the ClassicalFG.lhs implementation.

> buildChorale g [(cons, x)] km = 
>     snd $ snd $ classicalFGR g km x cons
> buildChorale g [(cons1, a), (cons2, b)] km = 
>     let [g1, g2, g3, g4, g5] = take 5 $ splitN g
>         aChords = snd $ classicalCS g1 km a cons1
>         bChords = snd $ classicalCS g2 km b cons2
>         partA = snd $ snd $ classicalFG' g3 aChords
>         partA' = snd $ snd $ classicalFG' g4 aChords
>         partB = snd $ snd $ classicalFG' g5 bChords
>     in  partA :+: partA' :+: partB :+: partA

A "jazz chorale" ads an extra step in the foreground generation, converting 
numerals to jazz chords before running the classical algorithms.

> buildJChorale g [(cons, x)] (k,m) =  
>     let (g1,g2) = split g
>         jChords = atTrans k $ snd $ jazzChords g1 x cons
>     in  snd $ snd $ classicalFG' g2 jChords
> buildJChorale g [(cons1, a), (cons2, b)] (k,m) = 
>     let [g1, g2, g3, g4, g5] = take 5 $ splitN g
>         aChords = atTrans k $ snd $ jazzChords g1 a cons1
>         bChords = atTrans k $ snd $ jazzChords g2 b cons2
>         partA = snd $ snd $ classicalFG' g3 aChords
>         partA' = snd $ snd $ classicalFG' g4 aChords
>         partB = snd $ snd $ classicalFG' g5 bChords
>     in  partA :+: partA' :+: partB :+: partA


A "weird chorale" is one where numerals are run through OPTC-space before
applying a classical foreground.

> qOPTC = satbR (mkStdGen 123) satbFilter2 optcEq 

Note: the key will not affect weird chorales due to the use 
of OPTC-equivalence.

> buildWChorale g [(cons, x)] km = 
>     let (g1, g2) = split g
>         optChords = toOPTC g1 x km
>     in  snd $ snd $ classicalFG' g2 optChords
> buildWChorale g [(cons1, a), (cons2, b)] km = 
>     let [g1, g2, g3, g4, g5] = take 5 $ splitN g
>         aChords = toOPTC g1 a km
>         bChords = toOPTC g2 b km
>         partA = snd $ snd $ classicalFG' g3 aChords
>         partA' = snd $ snd $ classicalFG' g4 aChords
>         partB = snd $ snd $ classicalFG' g5 bChords
>     in  partA :+: partA' :+: partB :+: partA

> toOPTC g x (k,m) = 
>     let aChords = atTrans k $ map toAbsChord x
>         es = map (eqClass qOPTC optcEq) $ map thd aChords
>     in  zipWith newP aChords $ greedyProg' vl7 nearFall g es

Jazz foregrounds are created using the two algorithms in JazzFG.lhs.

> buildJazzChords = buildJazz jazzFG1
> buildBossaNova = buildJazz jazzFG2

> buildJazz f g [(cons, x)] (k,m) = snd $ f g (ctTrans k x) []
> buildJazz f g [(cons1, a), (cons2, b)] (k,m) = 
>     let [g1,g2,g3] = take 3 $ splitN g
>         jA = snd $ f g1 (ctTrans k a) []
>         jA' = snd $ f g2 (ctTrans k a) []
>         jB = snd $ f g3 (ctTrans k b) []
>     in  jA :+: jA' :+: jB :+: jA