Simple Jazz Foreground Algorithms
Donya Quick

Last Modified 24-Apr-2013

> module JazzFG where
> import GrammarMonad hiding (c)
> import GrammarRules
> import GrammarRulesV
> import System.Random
> import Euterpea hiding (c)
> import ChordSpaces
> import PostProc
> import Data.List
> import FGTypes
> import ModeSpace
> import Control.Parallel.Strategies 
> import ClassicalFG 

> 

First, we find the modes for Roman numerals interpreted
in a particular key/mode.

> majorModes = allModes
> minorModes = drop 5 allModes ++ (take 5 allModes)

> chordMode :: CType -> Key -> AbsMode
> chordMode ct (k,m) = 
>     let pModes = if m==Major then majorModes else minorModes
>         ctMode = pModes !! fromEnum ct
>         ck = pModes !! 0 !! fromEnum ct
>     in  t (k+ck) ctMode

> toJTriple :: (Key, Dur, CType) -> (Key, Dur, AbsMode)
> toJTriple (km,d,c) = (km, d, chordMode c km)

> type JNote = (Key, Dur, AbsPitch)
> type JVoice = [JNote]

> toJNote :: (Key, Dur, [AbsPitch]) -> [JNote]
> toJNote (km, d, c) = map (\p -> (km,d,p)) c

> dtrans = Data.List.transpose

> jnToMusic :: [InstrumentName] -> [JVoice] -> Music Pitch
> jnToMusic is jvs = chord $ zipWith jnToMusic1 is jvs 

> jnToMusic1 :: InstrumentName -> JVoice -> Music Pitch 
> jnToMusic1 i v = instrument i $ line $ map f v where
>     f (km, d, p) = note d $ pitch p

============================


Algorithm 1: chords and a stochastic bass. Let instantiation only takes 
place at the level of Roman numerals. 

> --            r t f 7    r 2 t f 7	r=root, r=third, f=fifth
> alg1Temps = [[0,2,4,6], [0,1,2,4,6]] 

> --          bass       chords
> alg1Rans = (34,45) : take 4 (repeat (50,64))

> bassRoot (chrd, m) = (minimum chrd `mod` 12) == head (normO m)

> jazzFG1 :: StdGen -> [(Key, Dur, CType)] -> (StdGen, Music Pitch)
> jazzFG1 g chords = 
>     let [gJ, gR, gOPC, gB] = take 4 $ splitN g
>         jts = map toJTriple chords
>         ms = map (\(a,b,c) -> ([],c)) jts -- get just modes as JChords
>         qJ = modeSpace' alg1Temps -- subset of ModeSpace desired
>         chordsJ = greedyProg qJ modeEq (const True) nearFallJ gJ ms -- random walk through qj
>         qOPC = makeRange' alg1Rans // opcEq  -- subset of OPC-space desired
>         es = map (convOPC qOPC bassRoot) chordsJ -- OPC equivalence classes for chords
>         chordsOPC = greedyProg' (const True) nearFall gOPC es -- random walk through OPC-space
>         chordsOPC' = zipWith (\(a,b,c) x -> (a,b,x)) jts chordsOPC -- tag with dur & mode
>         jVoices = dtrans $ map toJNote chordsOPC' -- place in voice format
>         (gRet, bassLine) = stochBass gB $ head jVoices -- stochastic bassline
>     in  (gRet, instrument AcousticBass bassLine :=: 
>                jnToMusic (repeat AcousticGrandPiano) (tail jVoices))

> splitN g = let (g1,g2) = split g in g1 : splitN g2

Direct interface to grammar monad:

> jazzFG1T :: StdGen -> Key -> Term CType-> (StdGen, Music Pitch)
> jazzFG1T g (k,m) t = jazzFG1 g $ ctTrans k $ toChords (expand [] t) m

> convOPC :: QSpace AbsChord -> Predicate JChord -> JChord -> EqClass AbsChord
> convOPC q pj (c,m) = filter (\x -> pj (x,m)) $ eqClass q opcEq c

> stochBass :: StdGen -> [JNote] -> (StdGen, Music Pitch)
> stochBass g [] = (g, rest 0)
> stochBass g ((km,d,p):t) = 
>     let (g', pat) = pickPattern g d p
>         (g'', t') = stochBass g' t
>     in  (g'', pat :+: t')

> pickPattern g d p = 
>     let (r,g') = randomR (0,length pats - 1) g
>         f d p = note d (pitch p)
>         pats = [f d p,
>                 if d>=hn then f qn p :+: f (d-qn) p else f d p,
>                 if d>=hn then f (d-en) p :+: f en p else f d p]
>     in  (g', pats !! r)


=============================

Algorithm 2: simple bossa nova

This approach interprets Roman numerals through three separate
chord spaces in order to cut down the task's combinatorics. 

> alg2TempsC = [[0,2,4,6], [1,2,4,6]] -- for chords
> alg2TempsB = [[0,4]] -- for bass
> alg2TempsL = [[0],[2],[4]] -- for lead

> alg2RansB = [(34,49), (34,49)]
> alg2RansC = take 4 $ repeat (50,64)
> alg2RansL = [(65,80)]

> bassRoot2 ([b1,b2], m) = normO [b1,b2] == normO [m !! 0, m!! 4]
> bassRoot2 _ = error "(bassRoot2) Bad arguments."

> alg2FilterC x = sorted x && pianoChord x

> jazzFG2 :: StdGen -> [(Key, Dur, CType)] -> (StdGen, Music Pitch)
> jazzFG2 g chords = 
>     let gs@[gJC, gJB, gJL, gRC, gRB, gRL, gOPC_C, gOPC_B, gOPC_L, gL] = take 10 $ splitN g
>         jts = map toJTriple chords
>         ms = map (\(a,b,c) -> ([],c)) jts
>         qs@[qJC, qJB, qJL] = map modeSpace' [alg2TempsC, alg2TempsB, alg2TempsL] -- jazz spaces
>         [chordsJ, bassJ, leadJ] = -- random walk for chords
>             zipWith (\q gx -> greedyProg q modeEq (const True) nearFallJ gx ms) qs $
>             take 3 gs 
>         qOPC_C = filter alg2FilterC (makeRange' alg2RansC) // opcEq 
>         qOPC_B = makeRange alg2RansB // opcEq 
>         qOPC_L = makeRange' alg2RansL // opcEq 
>         esC = map (convOPC qOPC_C (const True)) chordsJ -- OPC equivalence classes for chords
>         esB = map (convOPC qOPC_B bassRoot2) bassJ
>         esL = map (convOPC qOPC_L (const True)) leadJ
>         chordsOPC = greedyProg' (const True) nearFall gOPC_C esC -- random walk through OPC-space
>         bassOPC = greedyProg' (noCPL 7) nearFall gOPC_B esB -- random walk for bass
>         leadOPC = greedyProg' (noCPL 7) nearFall gOPC_L esL -- random walk for lead
>         [cc, bc, lc] = map (zipWith (\(a,b,c) x -> (a,b,x)) jts) [chordsOPC, bassOPC, leadOPC] -- tag with dur & mode
>         cm = bossaChords cc
>         bm = bossaBass bc
>         (gRet, lm) = bossaLead gL lc
>     in  (gRet, chord [instrument AcousticBass bm,
>                       instrument AcousticGrandPiano cm, 
>                       instrument Flute lm])

> jazzFG2T :: StdGen -> Key -> Term CType -> (StdGen, Music Pitch)
> jazzFG2T g (k,m) t = jazzFG2 g $ ctTrans k $ toChords (expand [] t) m

> bossaBass :: [TChord] -> Music Pitch
> bossaBass [] = rest 0
> bossaBass ((km,d,c@[p1,p2]):t) = 
>     if d > wn then bossaBass ((km,wn,c):(km,d-wn,c):t) else
>     if d == wn then f1 p1 p2 :+: bossaBass t else
>     if d == hn then f2 p1 p2 :+: bossaBass t  else f3 p1 d :+: bossaBass t where
>     f1 b1 b2 = f2 b1 b2 :+: f2 b2 b1
>     f2 b1 b2 = f3 b1 (qn+en) :+: f3 b2 en 
>     f3 b1 d = note d (pitch b1)
> bossaBass _ = error "(bossaBass) Bad input"

> bossaChords :: [TChord] -> Music Pitch
> bossaChords [] = rest 0
> bossaChords ((km,d,c):t) = 
>     if d > wn then bossaChords ((km,wn,c):(km,d-wn,c):t) else 
>     if d==wn then f1 c :+: bossaChords t else f2 d c :+: bossaChords t where
>     f1 c = let c' = f2 en c in rest qn :+: c' :+: rest qn :+: c' :+: rest qn
>     f2 d c = chord $ map (\p -> note d $ pitch p) c


> toTN2' (k,d,[p]) b l = TNote2 k d b l p
> toTN2' _ _ _ = error "(toTN2') Bad arguments"

> tn2M (TNote2 k d b l p) = note d (pitch p)

> bossaLead :: StdGen -> [TChord] -> (StdGen, Music Pitch)
> bossaLead g ts = 
>     let ls = take (length ts - 1) (repeat False) ++ [True]
>         v = zipWith3 toTN2' ts (repeat 0) ls
>         (g', v') = addFgToVoice jConsts (foreFunsJ defConsts) g v
>     in  (g', line $ map tn2M v') where
>     foreFunsJ c = [(0.5, f1 c), (0.5, f2 c)] :: [(Double, ForeFun)]
>     jConsts = CConstants 2 3 0.3 0.5 0.8 7


