Chord Spaces Implementation
Donya Quick and Paul Hudak
Last modified: 14-June-2013
For paper: Grammar-Based Automated Music Composition in Haskell

This module is based on an earlier version that was for the paper
"Computing with Chord Spaces" (ICMC 2012). This version is part of
the implementation for "A Temporal Generative Graph Grammar for 
Harmonic and Metrical Structure" (ICMC 2013). Changes since the
2012 version include:
- Renamed type Chord to be AbsChord to avoid type conflicts with other modules.
- Simlification of existing code.
- Removal of the greedy method for traversing chord spaces (not 
  relevant to the TGGG/PTGG).
- Support for Octave, Permuatation, and Cardinality (OPC) 
  equivalence. 

> module ChordSpaces where
> import Data.List
> import System.Random

Type definitions:

> type Norm a = a -> a -- normalizations
> type EqRel a = a -> a -> Bool -- equivalence relations
> type AbsChord = [Int]
> type Prog = [AbsChord] -- Chord progression
> type EqClass a = [a] -- equivalence class
> type QSpace a = [EqClass a] -- quotient space
> type Predicate a = a -> Bool

The makeRange function will generate Z^n for user-specified ranges.

> makeRange :: [(Int, Int)] -> [[Int]]
> makeRange = foldr (\(l,u) xs -> [(a:b) | a<-[l..u], b<-xs]) [[]]


========= O, P, & T IMPLEMENTATION =========

First we will define the octave and transposition operations. 
For f(x)=y with f in {o, t, p}, x~y for the corresponding 
equivalence relation (O, T, and P respectively).

> o = zipWith (\i x -> x + 12 * i)
> t c = map (+c)

We define normalizations for O, P, T, OP, OT, and PT.
We also add a new definition, OPC.

> normO, normT, normP, normOP, normPT, normOPC :: Norm AbsChord
> normO = map (`mod` 12)
> normT x = map (subtract $ head x) x
> normP = sort
> normOP = sort . normO
> normPT = normT . sort
> normOT = normO . normT
> normOPC = nub . normOP

Given a normalization, it can be turned into an 
equivalence relation.

> normToEqRel :: (Eq a) => Norm a -> EqRel a
> normToEqRel f a b = f a == f b

> oEq, pEq, tEq, opEq, ptEq, opcEq :: EqRel AbsChord
> [oEq, pEq, tEq, opEq, ptEq, otEq, opcEq] = 
>     map normToEqRel [normO, normT, normP, normOP, normPT, normOT, normOPC]

For OPT equivalence, we make use of the following octStacks 
function that generates all octave stackings of a AbsChord, 
assuming that it is within [0,12]^n. 

> optEq :: EqRel AbsChord
> optEq a b = 
>     let (a', b') = (normT $ normOP a, normT $ normOP b)
>         octStacks x = zipWith o (makeRange $ take (length x) $ 
>                       repeat (0,1)) (repeat x)
>         s = map (normT . normP) $ octStacks b'
>     in  or (map (==a') s)



========= QUOTIENT SPACE IMPLEMENTATION =========

First we define the "slash" operator for S/R. 

> (//) :: (Eq a) => [a] -> EqRel a -> QSpace a
> [] // r = []
> xs // r = 
>     let sx = [y | y <- xs, r y (head xs)] 
>     in  sx : [z | z <- xs, not (elem z sx)] // r

The eqClass function is used to find the equivalence class of an
element, x, given a quotient space and the relation used to form it.
We need to know the relation used, because we do not require that
x is in the quotient space, qs.

> eqClass :: (Eq a, Show a) => QSpace a -> EqRel a -> a -> EqClass a
> eqClass qs r x = 
>     let ind = findIndex (\e -> r x (head e)) qs
>     in  maybe (error ("(eqClass) No class for "++show x)) (qs !!) ind 

The allSolns function finds all R-equivalent sequences given a quotient
space, qs, and a relation, r.



========= OBTAINING SOLUTIONS =========

This is the obvious, but slightly less efficient way to obtain solutions.
The lazy evaluation in Haskell allows it to do aggressive pruning similarly
to how pairProg handles it.

> allSolns :: (Eq a, Show a) => QSpace a -> EqRel a -> [a] -> [[a]]
> allSolns qs r [] = [[]]
> allSolns qs r (x:xs) = 
>     [(y:ys) | y <- eqClass qs r x, ys <- allSolns qs r xs]

Unfortunately, allSolns won't work with really big problems. Large chord 
spaces and long progressions (even with small chord spaces) cause an 
exponential blowup in the search space. Two alternative options below
prune the solution space more aggressively.

The pairProg function applies a predicate as it generates the 
solutions.

> pairProg :: (Eq a, Show a) => QSpace a -> EqRel a -> Predicate (a,a) -> [a] -> [[a]]
> pairProg qs r c [] = []
> pairProg qs r c [x] = map (\a -> [a]) $ eqClass qs r x
> pairProg qs r c (x:xs) = 
>     let endSolns = pairProg qs r c xs -- map xs to a new progression
>         f y soln = c (y, head soln) -- filter using the classifier
>         newSolns = [(y:ys) | y<-eqClass qs r x, ys<-endSolns, f y ys]
>     in  if null newSolns then error "No solutions that satisfy the consraints!"
>         else newSolns

Now we define the greedy algorithm, greedyProg.

We define choose, a function to stochastically select an element
from a list.

> choose :: StdGen -> [a] -> (StdGen, a)
> choose g xs = 
>     let (r, g') = next g
>     in  (g', xs !! (r `mod` length xs))

And finally the recursive, greedy function, greedyProg, and its
"helper," greedyChord.


> type Fallback a = EqClass a -> StdGen -> a -> (StdGen, a)
		 	 
> greedyProg :: (Eq a, Show a) => QSpace a -> EqRel a -> 
>     Predicate (a,a) -> Fallback a -> StdGen -> [a] -> [a]
> greedyProg qs r c f g [] = []
> greedyProg qs r c f g (x:xs) =
>     let e = eqClass qs r x
>         (g', y0) = choose g e -- randomly choose the first AbsChord
>     in  greedyRec qs r c f g y0 xs where 
>         greedyRec qs r c f g y pts = 
>             let (g', yi) = greedyChord (eqClass qs r (head pts)) y c f g
>             in  if null pts then [y] 
>                 else y : greedyRec qs r c f g' yi (tail pts)

> greedyChord :: (Eq a, Show a) => EqClass a -> a -> Predicate (a,a) -> 
>     Fallback a -> StdGen -> (StdGen, a)
> greedyChord e yprev hpair f g = 
>     let (rand, g') = next g
>         yxs = zip (repeat yprev) e
>         ys = map snd $ filter hpair yxs
>     in  if null ys then f e g' yprev -- fallback case
>         else (g', ys !! (rand `mod` length ys)) -- ok case


Now we need some fall-back functions. We can construct them from
predicates.

> fallFun :: (Show a) => Predicate (a,a) -> EqClass a -> StdGen -> a -> (StdGen, a)
> fallFun c e g x = 
>     let ys = map snd $ filter c $ zip (repeat x) e
>         (i, g') = next g
>     in  if null ys then error ("Stuck at symbol "++show x++". No viable options left!")
>         else (g', ys !! (i `mod` length ys))

This is the default fallback function we used in our experiments:

> defFall = fallFun (maxClass 7)


And also a nearest neighbor fallback function that would always 
succeed if the equivalence class has at least one element (which 
should always be the case).

> nearFall ::  EqClass AbsChord -> StdGen -> AbsChord -> (StdGen, AbsChord)
> nearFall e g x = 
>     let ds = map (simpleDist x) e :: [Double]
>         y = e !! (head $ findIndices (==minimum ds) ds)
>     in  (g, y)



======== PREDICATES =========

First we defint the hProg function that turns a pairwise predicate 
into a progression predicate.

> hProg :: Predicate (a,a) -> Predicate [a]
> hProg f xs = and $ map f $ zip xs $ tail xs

> foldPreds :: [Predicate a] -> Predicate a
> foldPreds fs xs = and $ map ($xs) fs

Here we define the notion of voices not crossing to mean that if there 
is a permutation that sorts both chords, the voices do not cross. This 
allows for voices that may be in unison.

> hNotCross1 :: Predicate (AbsChord, AbsChord)
> hNotCross1 (c1,c2) = 
>     let c1' = sort c1
>         ps1 = findAllP c1 (sort c1) -- find permutation, p1, to sort c1
>         ps2 = findAllP c2 (sort c2) -- find permutation, p2, to sort c2
>         ps = [p | p<-ps1, elem p ps2]
>     in  not $ null ps 


The function above needs to generate all permutations (as indices) 
of a chord.

> findAllP :: AbsChord -> AbsChord -> [[Int]]
> findAllP c1 c2 = 
>     let n = length c1
>         f i = filter (\j -> c2 !! j == c1 !! i) [0..n - 1]
>         g [] = [[]]
>         g (is:t) = [(a:b) | a<-is, b<-g t]
>     in  filter (\x -> x == nub x) (g $ map f [0..n - 1])

> hNotCross2 :: Predicate Prog
> hNotCross2 = hProg hNotCross1

Now we define a predicate for avoiding parallel motion (all voices 
moving in the same direction). We will consider parallel motion in 
the context of vectors, so this is not the strictest possible case.
We don't care about intervals of 0, since that does not constitute
voice movement.

> hNotPar1 :: Predicate (AbsChord, AbsChord)
> hNotPar1 (c1, c2) = 
>     let diffs = zipWith subtract c1 c2
>     in  not $ hasDups $ filter (/= 0) diffs

> hasDups :: (Eq a) => [a] -> Bool
> hasDups [] = False
> hasDups (a:as) = elem a as || hasDups as

Now we define a stricter version that considers parallel motion
in MULTISETS rather than vectors. On more than 2 voices, this 
will become rather hard to satisfy unless the chords are 
quite strange.

> hNotParStrict1 :: Predicate (AbsChord, AbsChord)
> hNotParStrict1 (c1,c2) = hNotPar1 (sort c1, sort c2)

And finally, progression level predicates of each.

> hNotPar2, hNotParStrict2 :: Predicate Prog
> hNotPar2 = hProg hNotPar1
> hNotParStrict2 = hProg hNotParStrict1


And now a "fill in the blanks" predicate. With this predicate,
we can set start and end points, middle points, etc. if we 
want them. Of course, there will be no valid solutions unless we
pick points within the intended quotient space.

> fillBlanks :: [Maybe AbsChord] -> Predicate Prog
> fillBlanks (m:ms) (p:ps) = 
>     case m of Just c -> p == c && fillBlanks ms ps
>               Nothing -> fillBlanks ms ps 
> fillBlanks [] [] = True
> fillBlanks _ _ = False


Now some distance metric-type approaches.

> type DistMeasure = AbsChord -> AbsChord -> Double
> type Threshold = Double

> simpleDist, eucDist, maxDist :: DistMeasure
> simpleDist a b = fromIntegral $ sum $ map abs $ zipWith subtract a b
> eucDist a b = sqrt $ sum $ map fromIntegral $ zipWith subtract a b
> maxDist a b = fromIntegral $ maximum $ map abs $ zipWith subtract a b

> distClass :: DistMeasure -> Predicate Double -> Predicate (AbsChord, AbsChord)
> distClass d ft (x,y) = ft $ d x y

> simpleClass t = distClass simpleDist (<=t)
> eucClass t = distClass eucDist (<=t)
> maxClass t = distClass maxDist (<=t)

> noCPL :: Double -> Predicate (AbsChord, AbsChord)
> noCPL i x = maxClass i x && hNotPar1 x && hNotCross1 x

> noCL :: Double -> Predicate (AbsChord, AbsChord)
> noCL i x = maxClass i x && hNotCross1 x

===============

> randomize :: StdGen -> [a] -> [a]
> randomize sg rs = 
>     let n = length rs
>         plist = take n (nub (randomRs (0,n-1) sg))
>     in  map (rs!!) plist